@@ -1,15 +1,19 @@
 import * as fs from 'fs'
-import * as glob from 'glob'
-import * as Path from 'path'
+import * as FuzzySearch from 'fuzzy-search'
 import * as request from 'request-promise-native'
 import * as URI from 'urijs'
+import { promisify } from 'util'
 import * as LSP from 'vscode-languageserver'
 import * as Parser from 'web-tree-sitter'
 
-import { uniqueBasedOnHash } from './util/array'
+import { getGlobPattern } from './config'
 import { flattenArray, flattenObjectValues } from './util/flatten'
+import { getFilePaths } from './util/fs'
+import { getShebang, isBashShebang } from './util/shebang'
 import * as TreeSitterUtil from './util/tree-sitter'
 
+const readFileAsync = promisify(fs.readFile)
+
 type Kinds = { [type: string]: LSP.SymbolKind }
 
 type Declarations = { [name: string]: LSP.SymbolInformation[] }
@@ -27,51 +31,67 @@ export default class Analyzer {
    * Initialize the Analyzer based on a connection to the client and an optional
    * root path.
    *
-   * If the rootPath is provided it will initialize all *.sh files it can find
-   * anywhere on that path.
+   * If the rootPath is provided it will initialize all shell files it can find
+   * anywhere on that path. This non-exhaustive glob is used to preload the parser.
    */
-  public static fromRoot({
+  public static async fromRoot({
     connection,
     rootPath,
     parser,
   }: {
     connection: LSP.Connection
-    rootPath: string | null
+    rootPath: LSP.InitializeParams['rootPath']
     parser: Parser
   }): Promise<Analyzer> {
-    // This happens if the users opens a single bash script without having the
-    // 'window' associated with a specific project.
-    if (!rootPath) {
-      return Promise.resolve(new Analyzer(parser))
+    const analyzer = new Analyzer(parser)
+
+    if (rootPath) {
+      const globPattern = getGlobPattern()
+      connection.console.log(
+        `Analyzing files matching glob "${globPattern}" inside ${rootPath}`,
+      )
+
+      const lookupStartTime = Date.now()
+      const getTimePassed = (): string =>
+        `${(Date.now() - lookupStartTime) / 1000} seconds`
+
+      let filePaths: string[] = []
+      try {
+        filePaths = await getFilePaths({ globPattern, rootPath })
+      } catch (error) {
+        connection.window.showWarningMessage(
+          `Failed to analyze bash files using the glob "${globPattern}". The experience will be degraded. Error: ${error.message}`,
+        )
+      }
+
+      // TODO: we could load all files without extensions: globPattern: '**/[^.]'
+
+      connection.console.log(
+        `Glob resolved with ${filePaths.length} files after ${getTimePassed()}`,
+      )
+
+      for (const filePath of filePaths) {
+        const uri = `file://${filePath}`
+        connection.console.log(`Analyzing ${uri}`)
+
+        try {
+          const fileContent = await readFileAsync(filePath, 'utf8')
+          const shebang = getShebang(fileContent)
+          if (shebang && !isBashShebang(shebang)) {
+            connection.console.log(`Skipping file ${uri} with shebang "${shebang}"`)
+            continue
+          }
+
+          analyzer.analyze(uri, LSP.TextDocument.create(uri, 'shell', 1, fileContent))
+        } catch (error) {
+          connection.console.warn(`Failed analyzing ${uri}. Error: ${error.message}`)
+        }
+      }
+
+      connection.console.log(`Analyzer finished after ${getTimePassed()}`)
     }
 
-    return new Promise((resolve, reject) => {
-      glob('**/*.sh', { cwd: rootPath }, (err, paths) => {
-        if (err != null) {
-          reject(err)
-        } else {
-          const analyzer = new Analyzer(parser)
-          paths.forEach(p => {
-            const absolute = Path.join(rootPath, p)
-            // only analyze files, glob pattern may match directories
-            if (fs.existsSync(absolute) && fs.lstatSync(absolute).isFile()) {
-              const uri = `file://${absolute}`
-              connection.console.log(`Analyzing ${uri}`)
-              analyzer.analyze(
-                uri,
-                LSP.TextDocument.create(
-                  uri,
-                  'shell',
-                  1,
-                  fs.readFileSync(absolute, 'utf8'),
-                ),
-              )
-            }
-          })
-          resolve(analyzer)
-        }
-      })
-    })
+    return analyzer
   }
 
   private parser: Parser
@@ -110,18 +130,28 @@ export default class Analyzer {
     return symbols.map(s => s.location)
   }
 
+  /**
+   * Find all the symbols matching the query using fuzzy search.
+   */
+  public search(query: string): LSP.SymbolInformation[] {
+    const searcher = new FuzzySearch(this.getAllSymbols(), ['name'], {
+      caseSensitive: true,
+    })
+    return searcher.search(query)
+  }
+
   public async getExplainshellDocumentation({
-    pos,
+    params,
     endpoint,
   }: {
-    pos: LSP.TextDocumentPositionParams
+    params: LSP.TextDocumentPositionParams
     endpoint: string
   }): Promise<any> {
     const leafNode = this.uriToTreeSitterTrees[
-      pos.textDocument.uri
+      params.textDocument.uri
     ].rootNode.descendantForPosition({
-      row: pos.position.line,
-      column: pos.position.character,
+      row: params.position.line,
+      column: params.position.character,
     })
 
     // explainshell needs the whole command, not just the "word" (tree-sitter
@@ -131,7 +161,14 @@ export default class Analyzer {
     // encounters newlines.
     const interestingNode = leafNode.type === 'word' ? leafNode.parent : leafNode
 
-    const cmd = this.uriToFileContent[pos.textDocument.uri].slice(
+    if (!interestingNode) {
+      return {
+        status: 'error',
+        message: 'no interestingNode found',
+      }
+    }
+
+    const cmd = this.uriToFileContent[params.textDocument.uri].slice(
       interestingNode.startIndex,
       interestingNode.endIndex,
     )
@@ -155,7 +192,7 @@ export default class Analyzer {
       return { ...response, status: 'error' }
     } else {
       const offsetOfMousePointerInCommand =
-        this.uriToTextDocument[pos.textDocument.uri].offsetAt(pos.position) -
+        this.uriToTextDocument[params.textDocument.uri].offsetAt(params.position) -
         interestingNode.startIndex
 
       const match = explainshellResponse.matches.find(
@@ -193,21 +230,23 @@ export default class Analyzer {
     const locations: LSP.Location[] = []
 
     TreeSitterUtil.forEach(tree.rootNode, n => {
-      let name: string = null
-      let rng: LSP.Range = null
+      let name: null | string = null
+      let range: null | LSP.Range = null
 
       if (TreeSitterUtil.isReference(n)) {
         const node = n.firstNamedChild || n
         name = contents.slice(node.startIndex, node.endIndex)
-        rng = TreeSitterUtil.range(node)
+        range = TreeSitterUtil.range(node)
       } else if (TreeSitterUtil.isDefinition(n)) {
         const namedNode = n.firstNamedChild
-        name = contents.slice(namedNode.startIndex, namedNode.endIndex)
-        rng = TreeSitterUtil.range(n.firstNamedChild)
+        if (namedNode) {
+          name = contents.slice(namedNode.startIndex, namedNode.endIndex)
+          range = TreeSitterUtil.range(namedNode)
+        }
       }
 
-      if (name === query) {
-        locations.push(LSP.Location.create(uri, rng))
+      if (name === query && range !== null) {
+        locations.push(LSP.Location.create(uri, range))
       }
     })
 
@@ -217,27 +256,27 @@ export default class Analyzer {
   /**
    * Find all symbol definitions in the given file.
    */
-  public findSymbols(uri: string): LSP.SymbolInformation[] {
+  public findSymbolsForFile({ uri }: { uri: string }): LSP.SymbolInformation[] {
     const declarationsInFile = this.uriToDeclarations[uri] || {}
     return flattenObjectValues(declarationsInFile)
   }
 
   /**
-   * Find unique symbol completions for the given file.
+   * Find symbol completions for the given word.
    */
-  public findSymbolCompletions(uri: string): LSP.CompletionItem[] {
-    const hashFunction = ({ name, kind }: LSP.SymbolInformation) => `${name}${kind}`
+  public findSymbolsMatchingWord({ word }: { word: string }): LSP.SymbolInformation[] {
+    const symbols: LSP.SymbolInformation[] = []
 
-    return uniqueBasedOnHash(this.findSymbols(uri), hashFunction).map(
-      (symbol: LSP.SymbolInformation) => ({
-        label: symbol.name,
-        kind: this.symbolKindToCompletionKind(symbol.kind),
-        data: {
-          name: symbol.name,
-          type: 'function',
-        },
-      }),
-    )
+    Object.keys(this.uriToDeclarations).forEach(uri => {
+      const declarationsInFile = this.uriToDeclarations[uri] || {}
+      Object.keys(declarationsInFile).map(name => {
+        if (name.startsWith(word)) {
+          declarationsInFile[name].forEach(symbol => symbols.push(symbol))
+        }
+      })
+    })
+
+    return symbols
   }
 
   /**
@@ -271,16 +310,22 @@ export default class Analyzer {
         return
       } else if (TreeSitterUtil.isDefinition(n)) {
         const named = n.firstNamedChild
+
+        if (named === null) {
+          return
+        }
+
         const name = contents.slice(named.startIndex, named.endIndex)
         const namedDeclarations = this.uriToDeclarations[uri][name] || []
 
         const parent = TreeSitterUtil.findParent(n, p => p.type === 'function_definition')
-        const parentName = parent
-          ? contents.slice(
-              parent.firstNamedChild.startIndex,
-              parent.firstNamedChild.endIndex,
-            )
-          : null
+        const parentName =
+          parent && parent.firstNamedChild
+            ? contents.slice(
+                parent.firstNamedChild.startIndex,
+                parent.firstNamedChild.endIndex,
+              )
+            : '' // TODO: unsure what we should do here?
 
         namedDeclarations.push(
           LSP.SymbolInformation.create(
@@ -321,13 +366,25 @@ export default class Analyzer {
     const document = this.uriToTreeSitterTrees[uri]
     const contents = this.uriToFileContent[uri]
 
-    const node = document.rootNode.namedDescendantForPosition({ row: line, column })
+    if (!document.rootNode) {
+      // Check for lacking rootNode (due to failed parse?)
+      return null
+    }
+
+    const point = { row: line, column }
+
+    const node = TreeSitterUtil.namedLeafDescendantForPosition(point, document.rootNode)
+
+    if (!node) {
+      return null
+    }
 
     const start = node.startIndex
     const end = node.endIndex
     const name = contents.slice(start, end)
 
     // Hack. Might be a problem with the grammar.
+    // TODO: Document this with a test case
     if (name.endsWith('=')) {
       return name.slice(0, name.length - 1)
     }
@@ -335,55 +392,17 @@ export default class Analyzer {
     return name
   }
 
-  private symbolKindToCompletionKind(s: LSP.SymbolKind): LSP.CompletionItemKind {
-    switch (s) {
-      case LSP.SymbolKind.File:
-        return LSP.CompletionItemKind.File
-      case LSP.SymbolKind.Module:
-      case LSP.SymbolKind.Namespace:
-      case LSP.SymbolKind.Package:
-        return LSP.CompletionItemKind.Module
-      case LSP.SymbolKind.Class:
-        return LSP.CompletionItemKind.Class
-      case LSP.SymbolKind.Method:
-        return LSP.CompletionItemKind.Method
-      case LSP.SymbolKind.Property:
-        return LSP.CompletionItemKind.Property
-      case LSP.SymbolKind.Field:
-        return LSP.CompletionItemKind.Field
-      case LSP.SymbolKind.Constructor:
-        return LSP.CompletionItemKind.Constructor
-      case LSP.SymbolKind.Enum:
-        return LSP.CompletionItemKind.Enum
-      case LSP.SymbolKind.Interface:
-        return LSP.CompletionItemKind.Interface
-      case LSP.SymbolKind.Function:
-        return LSP.CompletionItemKind.Function
-      case LSP.SymbolKind.Variable:
-        return LSP.CompletionItemKind.Variable
-      case LSP.SymbolKind.Constant:
-        return LSP.CompletionItemKind.Constant
-      case LSP.SymbolKind.String:
-      case LSP.SymbolKind.Number:
-      case LSP.SymbolKind.Boolean:
-      case LSP.SymbolKind.Array:
-      case LSP.SymbolKind.Key:
-      case LSP.SymbolKind.Null:
-        return LSP.CompletionItemKind.Text
-      case LSP.SymbolKind.Object:
-        return LSP.CompletionItemKind.Module
-      case LSP.SymbolKind.EnumMember:
-        return LSP.CompletionItemKind.EnumMember
-      case LSP.SymbolKind.Struct:
-        return LSP.CompletionItemKind.Struct
-      case LSP.SymbolKind.Event:
-        return LSP.CompletionItemKind.Event
-      case LSP.SymbolKind.Operator:
-        return LSP.CompletionItemKind.Operator
-      case LSP.SymbolKind.TypeParameter:
-        return LSP.CompletionItemKind.TypeParameter
-      default:
-        return LSP.CompletionItemKind.Text
-    }
+  private getAllSymbols(): LSP.SymbolInformation[] {
+    // NOTE: this could be cached, it takes < 1 ms to generate for a project with 250 bash files...
+    const symbols: LSP.SymbolInformation[] = []
+
+    Object.keys(this.uriToDeclarations).forEach(uri => {
+      Object.keys(this.uriToDeclarations[uri]).forEach(name => {
+        const declarationNames = this.uriToDeclarations[uri][name] || []
+        declarationNames.forEach(d => symbols.push(d))
+      })
+    })
+
+    return symbols
   }
 }
